from .select import *
from pathlib import Path
import logging
from collections import Iterable

__doc__ = "The xml_file module takes care of all I/O interactions concerning xml files"

__all__ = ["load", "load_single", "load_these", "load_all", "write"]


def load(path):
    """
    Load(s) json file(s) and returns the dictionary/-ies
    Specifying a file_name: one file will be loaded.
    Specifying a directory: all `*.json` files will be loaded.

    Parameters
    ----------
    path : str
        path to a file_name or directory

    Returns
    -------
    dict
        dictionary representing the json ``{file_name: {data}}``

    """
    files = return_file_list_if_directory(path, file_ending=".xml", return_always_list=True)
    data = load_these(files)
    try:
        [value] = data.values()
        return value
    except ValueError:
        return data


def load_single(file_name):
    """
    Load a single xml file

    Parameters
    ----------
    file_name : str
        file_name to load from

    Returns
    -------
    dict
        the xml as ordered dict ``{collections.OrderedDict}``

    """
    if Path(file_name).is_dir():
        raise IsADirectoryError("given path is a directory not a file")

    from xmltodict import parse

    with open(Path(file_name), "r") as f:
        logging.info("loading file_name {}".format(file_name))
        f = f.read()
        return dict(parse(f))


def load_these(file_name_list):
    """
    Load specified xml files and return the data in a dictionary with file_name as key

    Parameters
    ----------
    file_name_list : Iterable
        list of file_names to load from

    Returns
    -------
    dict(collections.OrderedDict)
        the dictionaries from the files as values of file_name as key
        ``{file_name: {collections.OrderedDict}``

    """
    if not isinstance(file_name_list, Iterable):
        raise TypeError("Expected list, got {}".format(type(file_name_list)))

    data = dict()
    for file in file_name_list:
        data[file] = load_single(file)

    return data


def load_all(directory):
    """
    Load all xml files in the directory and return the data in a dictionary with file_name as key

    Parameters
    ----------
    directory : str
        the directory containing the xml files

    Returns
    -------
    dict(collections.OrderedDict)
        the dictionaries from the files as values of file_name as key
        ``{file_name: {collections.OrderedDict}}``
    """

    files = get_file_list_from_directory(directory, file_ending=".xml")
    data = load_these(files)

    return data


def _check_allowed_keys(data):
    from re import match
    if isinstance(data, dict):
        for key in data:
            if not match("^[a-zA-Z_]+[a-zA-Z0-9_.]*$", key):
                raise ValueError(f"key name is not allowed in xml: {key}")
            _check_allowed_keys(data[key])

    elif isinstance(data, (list, set, tuple)):
        for item in data:
            _check_allowed_keys(item)


def write(data, file_name, main_key_name=None):
    """
    Save xml file from dict or collections.OrderedDict to file

    Parameters
    ----------
    file_name : str
        the file_name to save under. if no ending is provided, saved as .xml
    data : dict, collections.OrderedDict
        the dictionary to be saved as xml
    main_key_name : str
        if the dict/OrderedDict does not have the main key as a single key present (``{main_element_name: dict}``),
        it needs to be specified
    Returns
    -------

    """
    if main_key_name:
        data = {main_key_name: data}

    if "." not in file_name:
        file_name += ".xml"

    logging.info(f"saving to file_name: {file_name}")

    if not check_file_name_ending(file_name, "xml"):
        logging.warning(
            f"file_name ending {file_name.split('.')[-1]} different to standard ({'xml'})"
        )

    _check_allowed_keys(data)

    from xmltodict import unparse

    with open(Path(file_name), "w+") as f:
        unparse(data, output=f)



from .select import *
from pathlib import Path
import logging
from collections import Iterable

__doc__ = (
    "The json_file module takes care of all I/O interactions concerning json files"
)
__all__ = ["load", "load_single", "load_these", "load_all", "write"]


def load(path):
    """
    Load(s) json file(s) and returns the dictionary/-ies
    Specifying a file_name: one file will be loaded.
    Specifying a directory: all `*.json` files will be loaded.

    Parameters
    ----------
    path : str
        path to a file_name or directory

    Returns
    -------
    dict
        dictionary representing the json ``{file_name: {data}}``

    """
    files = return_file_list_if_directory(path, file_ending=".json", return_always_list=True)
    data = load_these(files)
    try:
        [value] = data.values()
        return value
    except ValueError:
        return data


def load_single(file_name):
    """
    Load a single json file

    Parameters
    ----------
    file_name : str
        file_name to load from

    Returns
    -------
    dict
        the loaded json as a dict ``{data}``

    """
    from json import load

    if Path(file_name).is_dir():
        raise IsADirectoryError("given path is a directory not a file")

    with open(Path(file_name), "r") as f:
        logging.info("loading file_name {}".format(file_name))
        return load(f)


def load_these(file_name_list):
    """
    Load specified json files and return the data in a dictionary with file_name as key

    Parameters
    ----------
    file_name_list : Iterable
        list of file_names to load from

    Returns
    -------
    dict(dict)
        the dictionaries from the files as values of file_name as key
        ``{file_name: {data}}``

    """
    if not isinstance(file_name_list, Iterable):
        raise TypeError("Expected list, got {}".format(type(file_name_list)))

    data = dict()
    for file in file_name_list:
        data[file] = load_single(file)

    return data


def load_all(directory):
    """
    Load all json files in the directory and return the data in a dictionary with file_name as key

    Parameters
    ----------
    directory : str
        the directory containing the json files

    Returns
    -------
    dict(dict)
        the dictionaries from the files as values of file_name as key
        ``{file_name: {data}}``
    """

    files = get_file_list_from_directory(directory, file_ending=".json")
    data = load_these(files)

    return data


def write(data, file_name, beautify=True, sort=False):
    """
    Save json from dict to file

    Parameters
    ----------
    file_name : str
        the file_name to save under. if no ending is provided, saved as .json
    data : dict
        the dictionary to be saved as json
    beautify : bool, optional
        if the data is represented in single row or human readable presented (default: human readable)
    sort : bool, optional
        if the keys shall be ordered (default: false)

    """
    if "." not in file_name:
        file_name += ".json"

    if not check_file_name_ending(file_name, "json"):
        logging.warning(
            f"file_name ending {'.' + file_name.split('.')[-1]} different to standard ({'.json'})"
        )

    logging.info(f"saving to file_name: {file_name}")

    from json import dump

    with open(Path(file_name), "w") as fp:
        if beautify:
            if sort:
                dump(data, fp, indent=4, sort_keys=True)
            else:
                dump(data, fp, indent=4)
        else:
            if sort:
                dump(data, fp, sort_keys=True)
            else:
                dump(data, fp)



from .select import *
import csv
from pathlib import Path
import logging
from collections import Iterable

__doc__ = "The csv_file module takes care of all I/O interactions concerning csv files"


__all__ = [
    "load",
    "load_single",
    "load_these",
    "load_all",
    "write",
    "write_from_rows",
    "write_from_dict",
]


def _register_csv_dialect(**kwargs):
    """
    Register a csv dialect from kwargs with differences to the main unix dialect

    Parameters
    ----------
    kwargs : optional
        all parameters for changing from unix basic dialect

    """
    csv_dialect_options = {i for i in set(dir(csv.Dialect)) if "__" not in i}
    if not all(key in csv_dialect_options for key in kwargs.keys()):
        raise KeyError(
            f"only these keys for csv dialect are allowed: {csv_dialect_options}\nGiven keys: {kwargs.keys()}"
        )
    from os import name
    if "nt" == name:
        kwargs.update({"lineterminator": "\n"})

    csv.register_dialect("custom", **kwargs)


def load(path, **kwargs):
    """
    Load(s) csv file(s) and returns the rows
    Specifying a file_name: one file will be loaded.
    Specifying a directory: all `*.csv` files will be loaded.

    Parameters
    ----------
    path : str
        path to a file_name or directory
    kwargs : optional
        csv dialect options

    Returns
    -------
    list, dict
        list of lists if a single file_name was provided: ``[[row1.1, row1.2]]``
        dict of list of lists if multiple files provided: ``{file_name : [[row1.1, row1.2]]}``

    """
    files = return_file_list_if_directory(path, file_ending=".csv", return_always_list=True)
    data = load_these(files, **kwargs)
    try:
        [value] = data.values()
        return value
    except ValueError:
        return data


def load_single(file_name, **kwargs):
    """
    Load a csv file and return the rows

    Parameters
    ----------
    file_name : str
       file_name to load from
    kwargs : optional
        csv dialect options

    Returns
    -------
    list
        list of lists representing the csv data
        ``[[row1.1, row1.2]]``
    """
    if Path(file_name).is_dir():
        raise IsADirectoryError("given path is a directory not a file")

    if kwargs and "dialect" in kwargs:
        dialect = kwargs["dialect"]
    elif kwargs:
        _register_csv_dialect(**kwargs)
        dialect = "custom"
    else:
        dialect = "unix"

    with open(Path(file_name), "r") as f:
        data = list()
        rows = csv.reader(f, dialect=dialect)
        for row in rows:
            data.append(row)

    return data


def load_these(file_name_list, **kwargs):
    """
    Load specified csv files and return the rows in a dictionary with file_name as key

    Parameters
    ----------
    file_name_list : Iterable
        list of file_names to load from
    kwargs : optional
        csv dialect options

    Returns
    -------
    dict
        the rows from the files as values of file_name as key
        ``{file_name : [[row1.1, row1.2]]}``

    """
    if kwargs and "dialect" not in kwargs:
        _register_csv_dialect(**kwargs)

    if not isinstance(file_name_list, Iterable):
        raise TypeError("Expected list, got {}".format(type(file_name_list)))

    data = dict()
    for file in file_name_list:
        data[file] = load_single(file, **kwargs)

    return data


def load_all(directory, **kwargs):
    """
    Load all csv files in the directory and return the rows in a dictionary with file_name as key

    Parameters
    ----------
    directory : str
        the directory containing the csv files
    kwargs : optional
     csv dialect options

    Returns
    -------
    dict
        the rows from the files as values of file_name as key
        ``{file_name : [[row1.1, row1.2]]}``
    """

    files = get_file_list_from_directory(directory, file_ending=".csv")
    data = load_these(files, **kwargs)

    return data


def write_from_rows(rows, file_name, **kwargs):
    """
    Save row based document from rows to file

    Parameters
    ----------
    file_name : str
        the file_name to save the data under. if no ending is provided, saved as `file_name.csv`
    rows : list
        list of lists to write to file_name
    kwargs : optional
        csv dialect options

    """
    if "." not in file_name.__str__():
        if isinstance(file_name, str):
            file_name += ".csv"
        elif isinstance(file_name, Path):
            file_name = file_name.joinpath(".csv")
        else:
            raise TypeError("unsupported type for file_name")

    logging.info(f"saving to file_name: {file_name}")

    if kwargs and "dialect" in kwargs:
        from os import name
        if "nt" == name:
            dialect = {i: csv.get_dialect(kwargs["dialect"]).__getattribute__(i) for i in
                       dir(csv.get_dialect(kwargs["dialect"])) if not i.startswith("__") and i != "strict"}
            dialect.update({"lineterminator": "\n"})
            _register_csv_dialect(**dialect)
            dialect = "custom"
        else:
            dialect = kwargs["dialect"]
    elif kwargs:
        _register_csv_dialect(**kwargs)
        dialect = "custom"
    else:
        dialect = "unix"

    if not check_file_name_ending(file_name, ["csv", "tsv"]):
        logging.warning(
            "file_name ending {} different to standard ({})".format(
                Path(file_name).parts[-1], ["csv", "tsv"]
            )
        )

    with open(Path(file_name), "w") as fw:
        w = csv.writer(fw, dialect=dialect)
        for row in rows:
            w.writerow(row)


def write_from_dict(
    data,
    file_name,
    main_key_name=None,
    main_key_position=0,
    order=None,
    if_empty_value=None,
    **kwargs,
):
    """
    Save a row based document from dict to file

    Parameters
    ----------
    file_name : str
        the file_name to save under. if no ending is provided, saved as `file_name.csv`
    data : dict
        the dictionary to be saved as csv
    main_key_name : str, optional
        if the json or dict does not have the main key as a single key present (``{main_element_name: dict}``),
        it needs to be specified
    order : dict {int: str}, list, optional
        for defining a specific order of the keys
        either a dictionary with the specified positions in a dictionary with positions as keys (integers) or in a list
    if_empty_value : any, optional
        the value to set when no handling is available
        default is "delete" leading to be an empty value
    main_key_position : int, optional
        the position in csv of the dictionary main key
    kwargs : optional
        csv dialect options

    """
    raise NotImplemented("bug in datesy")
    if not main_key_name:
        from datesy.inspect import cast_main_key
        data, main_key_name = cast_main_key(data)

    from datesy.convert import dict_to_rows

    rows = dict_to_rows(
        data=data,
        main_key_name=main_key_name,
        main_key_position=main_key_position,
        if_empty_value=if_empty_value,
        order=order,
    )
    write_from_rows(rows, file_name, **kwargs)


def write(
    data,
    file_name,
    main_key_name=None,
    main_key_position=0,
    order=None,
    if_empty_value=None,
    **kwargs,
):
    """
    Save a row based document from dict or list to file
    If presented a dictionary, converting to rows is done by the
    `dict_to_rows <https://datesy.readthedocs.io/en/latest/datesy.html#datesy.convert.rows_to_dict>`_
    method.

    Parameters
    ----------
    file_name : str
        the file_name to save under. if no ending is provided, saved as `file_name.csv`
    data : dict, list
        the dictionary or list to be saved as csv
    main_key_name : str, optional
        if the json or dict does not have the main key as a single key present (``{main_element_name: dict}``),
        it needs to be specified
    main_key_position : int, optional
        the position in csv of the dictionary main key
    order : dict, list, optional
        for defining a specific order of the keys. if dict, format: ``{int: str}``
        either a dictionary with the specified positions in a dictionary with positions as keys (integers) or in a list
    if_empty_value : any, optional
        the value to set when no handling is available
        default is "delete" leading to be an empty value
    kwargs : optional
        csv dialect options
    """
    if isinstance(data, list):
        if main_key_name or order or if_empty_value or main_key_position:
            raise ValueError(
                "if row of rows used, main_key_name, order, "
                "if_empty_value and main_key_position must not be set"
            )
        write_from_rows(data, file_name, **kwargs)
    elif isinstance(data, dict):
        write_from_dict(
            data,
            file_name,
            main_key_name,
            order,
            if_empty_value,
            main_key_position,
            **kwargs,
        )
    else:
        raise TypeError(
            "wrong type for `handling`. only list or dict are allowed, {} given".format(
                type(data)
            )
        )



from pandas import read_excel, ExcelFile
from .select import *
from openpyxl.styles import Alignment
import logging
import re
from pathlib import Path
from collections import Iterable

__doc__ = (
    "The xls_file module takes care of all I/O interactions concerning xls(x) files"
)

__all__ = [
    "load_single_sheet",
    "load_these_sheets",
    "load_all_sheets",
    "load_these_files",
    "load_all_files",
    "load",
    "write_single_sheet_from_DataFrame",
    "write_multi_sheet_from_DataFrames",
    "write_single_sheet_from_dict",
    "write_multi_sheet_from_dict_of_dicts",
]


def load(path):
    """
    Load all xls(x) files in the directory with all their sheets to
    a pandas.DataFrame as values to sheet_names as keys in a dictionary
    Specifying a file_name: one file will be loaded.
    Specifying a directory: all `*.xls(x)` files will be loaded.

    Parameters
    ----------
    path : str
        path to a file_name or directory

    Returns
    -------
    dict
        dictionary containing the sheets as `panda.DataFrames`: ``{file_name: {sheet_name: pandas.DataFrame}}``

    """
    files = return_file_list_if_directory(path, pattern="*.xls*", return_always_list=True)
    data = load_these_files(files)
    try:
        [value] = data.values()
        return value
    except ValueError:
        return data


def load_single_sheet(file_name, sheet=None):
    """
    Load a xls(x) file's (first) sheet to a pandas.DataFrame

    Parameters
    ----------
    file_name : str
       file_name to load from
    sheet : str, optional
        a specified sheet_name to extract. default is first sheet

    Returns
    -------
    pandas.DataFrame
        pandas.DataFrame representing the xls(x) file
    """
    file_name = Path(file_name)

    if file_name.is_dir():
        raise IsADirectoryError("given path is a directory not a file")

    elif not file_name.is_file():
        raise FileNotFoundError("given path doesn't point to a file")

    if not sheet:
        data = read_excel(file_name)
    else:
        data = read_excel(file_name, sheet_name=sheet)

    return data


def load_all_sheets(file_name):
    """
    Load from a xls(x) file all its sheets to a pandas.DataFrame as values to sheet_names as keys in a dictionary

    Parameters
    ----------
    file_name : str
       file_name to load from

    Returns
    -------
    dict
        dictionary containing the sheet_names as keys and pandas.DataFrame representing the xls(x) sheets
        ``{sheet_name: pandas.DataFrame}``
    """
    file_name = Path(file_name)
    excel_file = ExcelFile(file_name)
    return load_these_sheets(file_name, list(excel_file.sheet_names))


def load_these_sheets(file_name, sheets):
    """
    Load from a xls(x) file_name the specified sheets to a pandas.DataFrame
    as values to sheet_names as keys in a dictionary

    Parameters
    ----------
    file_name : str
       file_name to load from
    sheets : list
        sheet_names to load

    Returns
    -------
    dict(pandas.DataFrame)
        dictionary containing the sheet_names as keys and pandas.DataFrame representing the xls(x) sheets
        ``{sheet_name: pandas.DataFrame}``
    """

    data = dict()
    for sheet in sheets:
        data[sheet] = load_single_sheet(file_name, sheet)

    return data


def load_these_files(file_name_list):
    """
    Load the specified xls(x) files with all their sheets to
    a pandas.DataFrame as values to sheet_names as keys in a dictionary

    Parameters
    ----------
    file_name_list : Iterable
        list of file_names to load from

    Returns
    -------
    dict
        the data from the sheets in a dictionary with sheet_name as key within again a dictionary with file_name as key
        ``{file_name: {sheet_name: pandas.DataFrame}}``
    """
    if not isinstance(file_name_list, Iterable):
        raise TypeError(f"Expected list, got {type(file_name_list)}")

    data = dict()
    for file in file_name_list:
        data[file] = load_all_sheets(file)

    return data


def load_all_files(directory):
    """
    Load all xls(x) files in the directory with all their sheets to a pandas.DataFrame
    as values to sheet_names as keys in a dictionary

    Parameters
    ----------
    directory : str
        the directory containing the xlsx files

    Returns
    -------
    dict
        the data from the sheets in a dictionary with sheet_name as key within again a dictionary with file_name as key
        ``{file_name: {sheet_name: pandas.DataFrame}}``

    """

    files = get_file_list_from_directory(directory, pattern="*.xls*")
    data = load_these_files(files)

    return data


def __write_xlsx(data, file_name, auto_size_cells=True):
    """
    Write xlsx sheets from list of tuples ``[(sheet_name, DataFrame)]``

    Parameters
    ----------
    file_name : str
        the file_name to save the data under. if no ending is provided, saved as `file_name.xlsx`
    data : list
        list of tuples containing the sheet_name (pos.1) and panda.DataFrames (pos.2)
        ``[(sheet_name1, pandas.DataFrame1), sheet_name2, pandas.DataFrame2]``
    auto_size_cells : bool, optional
        if the auto-sizing of the cells shall be active

    """
    from pandas import ExcelWriter, DataFrame

    if not isinstance(file_name, str):
        raise TypeError(f"file_name needs to be a string. {type(file_name)} given")
    if not all(isinstance(element[1], DataFrame) for element in data):
        raise TypeError("all handling elements[1] need to be a pandas.DataFrame")

    if "." not in file_name:
        file_name += ".xlsx"

    with ExcelWriter(Path(file_name)) as writer:
        logging.info(f"saving to file_name: {file_name}")
        for element in data:
            element[1].to_excel(writer, sheet_name=element[0])

        # auto sizing the cells
        if auto_size_cells:
            for sheet in writer.sheets:
                worksheet = writer.sheets[sheet]
                for column in worksheet.columns:
                    # only take unmerged cells into account
                    unmerged_cells = list(
                        filter(
                            lambda cell_to_check: cell_to_check.coordinate
                            not in worksheet.merged_cells,
                            column,
                        )
                    )
                    # get length of longest line
                    length = max(
                        len(max(re.split("[\n|\r]", str(cell.value))))
                        for cell in unmerged_cells
                    )
                    # set the length of each column
                    worksheet.column_dimensions[
                        unmerged_cells[0].column_letter
                    ].width = (length + 3)
                    # multi-line cells: automatic sizing of height
                    for cell in unmerged_cells:
                        cell.alignment = Alignment(wrapText=True)

        writer.save()


def write_single_sheet_from_DataFrame(
    data_frame, file_name, sheet_name=None, auto_size_cells=True
):
    """
    Save a pandas.DataFrame to file

    Parameters
    ----------
    file_name : str
        the file_name to save under. if no ending is provided, saved as .xlsx
    data_frame : pandas.DataFrame
        pandas.DataFrame to write to file_name
    sheet_name : str, optional
        a sheet_name containing the data
    auto_size_cells : bool, optional
        if the auto-sizing of the cells shall be active

    """
    if not sheet_name:
        sheet_name = "Sheet1"

    __write_xlsx([(sheet_name, data_frame)], file_name, auto_size_cells)


def write_multi_sheet_from_DataFrames(
    data_frames, file_name, sheet_order=None, auto_size_cells=True
):
    """
    Save multiple pandas.DataFrames to one file

    Parameters
    ----------
    file_name : str
        the file_name to save under. if no ending is provided, saved as .xlsx
    data_frames : dict {sheet_name: DataFrame}
        dict of data_frames
    sheet_order : dict {int: str}, list, optional
        either a dictionary with the specified positions in a dictionary with positions as keys (integers) or in a list
    auto_size_cells : bool, optional
        if the auto-sizing of the cells shall be active

    """
    if sheet_order:
        from datesy.sort import create_sorted_list_from_order

        order = create_sorted_list_from_order(
            sheet_order, all_elements=data_frames.keys()
        )
        __write_xlsx(
            [(key, data_frames[key]) for key in order], file_name, auto_size_cells
        )
    else:
        __write_xlsx(
            [(key, data_frames[key]) for key in data_frames], file_name, auto_size_cells
        )


def write_single_sheet_from_dict(
    data,
    file_name,
    main_key_name=None,
    sheet=None,
    order=None,
    inverse=False,
    auto_size_cells=True,
):
    """
    Save a dictionary (``{main_key_name: {data}}``) as xlsx document to file
    Uses the `dict_to_pandas_data_frame
    <https://datesy.readthedocs.io/en/latest/datesy.html#datesy.convert.dict_to_pandas_data_frame>`_
    method for converting the dictionary to pandas.DataFrame.

    Parameters
    ----------
    file_name : str
        the file_name to save under. if no ending is provided, saved as .xlsx
    data : dict
        the dictionary to be saved as xlsx ``{main_key_name: {data}}``
    main_key_name : str, optional
        if the json or dict does not have the main key as a single `{main_element : dict}` present,
        it needs to be specified
    sheet : str, optional
        a sheet name for the handling
    order : dict, list, optional
        either a dictionary with the specified positions in a dictionary with positions as keys (integers) or in a list
    inverse : bool, optional
        if columns and rows shall be switched
    auto_size_cells : bool, optional
        if the auto-sizing of the cells shall be active

    """
    from datesy.convert import dict_to_pandas_data_frame

    data_frame = dict_to_pandas_data_frame(data, main_key_name, order, inverse)

    write_single_sheet_from_DataFrame(
        file_name=file_name,
        data_frame=data_frame,
        sheet_name=sheet,
        auto_size_cells=auto_size_cells,
    )


def write_multi_sheet_from_dict_of_dicts(
    data, file_name, order=None, auto_size_cells=True
):
    """
    Save dictionaries (``{sheet_name: {main_key_name: {data}}}``) as xlsx document to file
    Uses the `dict_to_pandas_data_frame
    <https://datesy.readthedocs.io/en/latest/datesy.html#datesy.convert.dict_to_pandas_data_frame>`_
    method for converting the dictionary to pandas.DataFrame.


    Parameters
    ----------
    file_name : str
        the file_name to save under. if no ending is provided, saved as .xlsx
    data : dict
        the dictionary to be saved as xlsx ``{sheet_name: {main_key_name: {data}}}``
    order : dict, list, optional
        either a dictionary with the specified positions in a dictionary with positions as keys (integers) or in a list
    auto_size_cells : bool, optional
        if the auto-sizing of the cells shall be active

    """

    from datesy.convert import dict_to_pandas_data_frame

    data_frames = {key: dict_to_pandas_data_frame(data[key]) for key in data}

    write_multi_sheet_from_DataFrames(
        data_frames, file_name, order, auto_size_cells=auto_size_cells
    )



import logging, os, glob, re
from pathlib import Path

__doc__ = "The file_selection module provides multiple supporting functions for interaction with files"

__all__ = [
    "get_newest_file_from_directory",
    "get_file_list_from_directory",
    "return_file_list_if_directory",
    "check_file_name_ending",
]


def return_file_list_if_directory(
    path, file_ending=None, pattern=None, regex=None, return_always_list=False
):
    """
    Return all files in directory (optionally specified with options) if path is a directory

    Parameters
    ----------
    path : str
        the path to test if directory
    file_ending : str, set, optional
        the file_name ending specifying the file_name type for the files in the directory
    pattern : str, optional
        pattern for the file_names in directory to match
        ``DataFile_*.json`` where ``*`` could be a date or other strings
    regex : str, optional
        a regular_expression (`regex <https://www.tutorialspoint.com/python/python_reg_expressions.htm>`_)
        for pattern matching of the file_names
    return_always_list : bool, optional
        if a single path shall be returned as in a list

    Returns
    -------
    list, str
        if directory the list of files else the path (in a list if `return_always_list` is set)

    """
    path = Path(path)
    if path.is_dir():
        return get_file_list_from_directory(path, file_ending, pattern, regex)
    if return_always_list:
        return [path]
    raise TypeError


def check_file_name_ending(file_name, ending):
    """
    Check if the file_name has the expected file_ending

    If one of the provided endings is the file_name's ending return `True`, else `False`

    Parameters
    ----------
    file_name : str
        The file_name to check the ending for
        The file_name may contain a path, so ``file_name.ending`` as well as ``path/to/file_name.ending`` will work

    ending : str, set, list
        The desired ending or multiple desired endings
        For single entries e.g. ``.json`` or ``csv``, for multiple endings e.g. ``['.json', 'csv']``

    Returns
    -------
    bool
        `True` if the `file_name`'s ending is in the given `ending`, else `False`

    """
    # input type check
    if not isinstance(file_name, (str, Path)):
        raise TypeError(f"file_name needs to be string, {type(file_name)} provided")
    if not isinstance(ending, (str, list, set, tuple)):
        raise TypeError(
            f"ending needs to be either a list, set, tuple or a string, {type(ending)} provided"
        )

    # check if multiple endings got provided
    if not hasattr(ending, "__iter__") or isinstance(ending, str):
        ending = [ending]
    elif isinstance(ending, (tuple, set)):
        ending = list(ending)

    # remove '.' from ending if provided as first character
    for element in ending:
        if not element.startswith("."):
            ending[ending.index(element)] = "." + element

    if Path(file_name).suffix in ending:
        return True

    return False


def get_file_list_from_directory(directory, file_ending=None, pattern=None, regex=None):
    """
    Return all files (optionally filtered) from directory in a list

    Parameters
    ----------
    directory : str
        the directory containing the desired files
    file_ending : str, set, optional
        the file_name's ending specifying the file type
    pattern : str, optional
        pattern for the file_names to match
        ``DataFile_*.json`` where ``*`` could be a date or other strings
    regex : str, optional
        a regular_expression (`regex <https://www.tutorialspoint.com/python/python_reg_expressions.htm>`_)
        for pattern matching

    Returns
    -------
    list
        a list of all relative file_name directories
    """
    # input type check
    if not isinstance(directory, (str, Path)):
        raise TypeError(
            "file_name needs to be string, {} provided".format(type(directory))
        )
    if pattern and regex:
        raise ValueError(
            "only `pattern` or `regex` may be specified"
        )
    if file_ending and not isinstance(file_ending, (str, list)):
        raise TypeError(
            "ending needs to be either a list or a string, {} provided".format(
                type(file_ending)
            )
        )
    if pattern and not isinstance(pattern, str):
        raise TypeError("pattern needs to be string, {} provided".format(type(pattern)))
    if regex and not isinstance(regex, str):
        raise TypeError("regex needs to be string, {} provided".format(type(regex)))

    directory = Path(directory)
    if not directory.is_dir():
        raise NotADirectoryError(f"{directory} is not a file_directory")

    directory_str = directory.__str__() + "/"

    # get list of files from directory
    if pattern:
        if "." not in pattern:  # if no file_name ending was specified in `pattern`
            files = glob.glob(directory_str + pattern + ".*")
        else:
            files = glob.glob(directory_str + pattern)
    else:
        files = glob.glob(directory_str + "*.*")

    # delete file_names not fitting the regex
    if regex:
        if file_ending and file_ending not in regex:
            if "." != file_ending[0]:
                file_ending = "." + file_ending

            if "$" == regex[-1]:
                regex = regex[:-1] + file_ending + "$"
            else:
                regex += file_ending + "$"

        logging.debug("regex for desired files: {}".format(regex))
        for file in files.copy():
            if not bool(re.search(regex, Path(file).parts[-1])):
                files.remove(file)

    # delete file_names not of specified file_ending
    if file_ending:
        logging.debug("file_endings for desired files: {}".format(file_ending))
        for file in files.copy():
            if not check_file_name_ending(file, file_ending):
                files.remove(file)

    logging.info("{} files in directory {}: {}".format(len(files), directory, files))
    return files


def get_newest_file_from_directory(
    directory, file_ending=None, pattern=None, regex=None
):
    """
    Return the latest file_name (optionally filtered) from a directory

    Parameters
    ----------
    directory : str
        the directory where to get the latest file_name from
    file_ending : str, set, optional
        the file_name ending specifying the file_name type
    pattern : str, optional
        pattern for the file_name to match
        ``DataFile_*.json`` where ``*`` could be a date or other strings
    regex : str, optional
       a regular_expression (`regex <https://www.tutorialspoint.com/python/python_reg_expressions.htm>`_)
       for pattern matching

    Returns
    -------
    str
        the `file_name` with the latest change date

    """

    files = get_file_list_from_directory(directory, file_ending, pattern, regex)
    if not files:
        raise ValueError(f"no files in directory found")
    latest_file = max(files, key=os.path.getctime)

    return latest_file



"""
The setup file_name.
If development mode (=changes in package code directly delivered to python) `pip install -e .` in directory of this file_name
"""

from setuptools import setup
from static_information import *

# https://python-packaging.readthedocs.io/en/latest/minimal.html

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name=project_name,
    version=version,
    description="basic tools for taking care for making the intro to handling handling with python easier",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=url,
    author=author,
    author_email=author_email,
    license=packet_license,
    packages=["fil_io"],
    python_requires=">=3.6",
    zip_safe=False,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
    ],
    # https://pypi.org/pypi?%3Aaction=list_classifiers
    install_requires=["pandas", "xmltodict", "openpyxl", "xlrd"],
)
